/* $Workfile: T8pBurstBasic.c $       $Revision: 11 $      $Modtime: 4/18/07 4:14p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8pBurstBasic.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

      Thor-2-PCI-Plus       (burst data looped via TSS)
      Thor-2-PCI-Express    (burst data looped via TSS)
      Thor-4-PCI-Plus       (burst data looped via TSS)
      Thor-8-PCI-Plus       (burst data looped via TSS)
      Thor-2-ExpressCard    (burst data looped via E1T1 devices and the DSP device)

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.
    
    It configures the burst device which is looped back to the burst device via
    the time-space-switch (if available) or via T1E1 devices.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxT1E1.h"  
#include "OtxDsp.h"  

#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "Dsp55Demo2Io.h"   // IO Control codes for the DSP to host communication
#include <Dsp55Ioctl.h>     // IO Control codes for the DSP to host communication


#define DEMO_DEFAULT_LOG_FILE   "T8pBurstBasic.log"
#define DEMO_DEFAULT_BOARD_NO   0

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int                    coHighways;
  int                    nHwyTimeslots;
  int                    nCurrentHwy;
  int                    boardNo;
  OTX_BOOL               bDisableBurstTx;
  OTX_BOOL               bDisableBurstRx;
  OTX_BOOL               bUseExternalLoop;
  OTX_BOOL               bUseDspLoop;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/

OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;
OTX_BOOL    g_bReadBurstStatsThreadActive = OTX_FALSE;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs
OTX_HANDLE  g_hUserApp = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE  g_hDsp = OTX_INVALID_HANDLE_VALUE;


struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hBurst;
} g_hPhysDevs;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;

  OTX_UINT8   sqGammaData[128];

};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)

  OTX_BOOL    bGammaTestMode;
  OTX_INT32   nGammaTestErrors[128];
  OTX_UINT8   sqGammaTestData[128];
};

struct {
  struct sBurstSender mBurstSender[2];        // one for each possible highway hosted by the Burst device
  struct sBurstReceiver mBurstReceiver[2];    // one for each possible highway hosted by the Burst device
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening Thor-2-PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_1_PCI_PLUS:
      DemoLogf("Opening Thor-1-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_EXPRESSCARD:
      DemoLogf("Opening Thor-2-ExpressCard board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

  // Optionally (mainly for testing, or for improving system performance for single direction burst applications,
  // disable one direction of the burst device
  if (g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
  if (g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
}




void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController, result);

  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    {
      // Loop the burst highway via the TSS device (since these boards all have a TSS device).
      OTX_INT32 iHighwayNo;
      OTX_HANDLE hPipe;
      OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      
      
      // Create the Pipes (one for each direction) between the Burst Device and the LIs
      for (iHighwayNo = 0; iHighwayNo < g_demoEnv.coHighways; iHighwayNo++) {
        
        memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));
        memset(sqBurstTimeSlotMask, 0xFF, g_demoEnv.nHwyTimeslots);
        
        DemoLogf("Creating pipe for Burst Sender/Receiver [highway=%d]\n", iHighwayNo);
        
        // --------------------------------------------------------------------
        // Create an incoming and outgoing pipe 
        // --------------------------------------------------------------------
        result = OtxDrvCreatePipe(64*g_demoEnv.nHwyTimeslots, &hPipe); DemoCheckRc(hPipe, result);
        
        // -------------------------
        // Connect pipe: Burst -> Burst
        // -------------------------
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nHwyTimeslots);
        DemoCheckRc(hPipe, result);
        
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nHwyTimeslots);
        DemoCheckRc(hPipe, result);
        
      }
    }
    break;
    
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {

      if (g_demoEnv.bUseDspLoop) {
        OTX_HANDLE hDualE1T1;
        Demo2IoInLoopAllTsS sIn;
        OTX_UINT32 nBytesReturned;
        OTX_UINT16 nDemo2Rc;
        char *pszDspProgram = "Dsp55Demo2NoExtMem.out";
        
#if 1 // This should not be needed. A bug in the 2B firmware?
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &hDualE1T1); DemoCheckRc(g_hPhysDevs.hThorController, result);
        result = OtxDrvEnable(hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(hDualE1T1, result);
#endif
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DSP, 0, 0, g_hEventQueue, &g_hDsp); DemoCheckRc(g_hPhysDevs.hThorController, result);

        // Get DSP Highway Speed
        OtxDrvGetAttributeValueUint32(g_hDsp, OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
        DemoLogf("DSP highway speed is %d kbps\n", g_nDspHighwaySpeed);
#if 0 // temporary
        DemoLogf("Disable g_hPhysDevs.hBurst\n");
        result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
#endif
        
        DemoLogf("Loading DSP Program: %s\n", pszDspProgram);
        result = OtxDspRunProgram(g_hDsp, pszDspProgram, "_c_int00"); DemoCheckRc(g_hDsp, result);
        
        result = OtxDrvCreateLogicalDevice(g_hDsp, OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, 0, &g_hUserApp); DemoCheckRc(g_hDsp, result);

        // Instruct the DSP program (Dsp55Demo2NoExtMem.out) to loop all timeslots
        sIn.bLoopAllTs = 1;
        
        result = OtxDspIoControl(
          g_hUserApp, 
          DEMO2_IOCTL_LOOP_ALL_TS, 
          &sIn, sizeof(sIn), 
          &nDemo2Rc, sizeof(nDemo2Rc), 
          &nBytesReturned, 
          OTX_TASK_SYNC, 
          5000 //OTX_TIME_INFINITY
          ); 
        DemoCheckRc(g_hUserApp, result);
        
      } else {
        // Loop the burst data in the T1E1 device device (62 of the 128 timeslots)
        OTX_HANDLE hDualE1T1;
        OTX_HANDLE hLi[2];
        OTX_INT32 i;
        OtxE1ConfigOptionsS sE1CfgData;
        OTX_UINT8 sqTimeSlotMask[32];
        OTX_HANDLE hPipe;

        // We'll source the Falc from the Burst device, 
        //  Falc <-- Burst Send (P2S)
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);

        // Open E1T1 devices
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &hDualE1T1); DemoCheckRc(g_hPhysDevs.hThorController, result);
        result = OtxDrvEnable(hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(hDualE1T1, result);
        
        for (i = 0; i < 2; i++) {
          result = OtxDrvOpenPhysicalDevice(hDualE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &hLi[i]); DemoCheckRc(hDualE1T1, result);
        }
        
        // Initialize E1 Configuration Data
        sE1CfgData.eTxLineCode                = OTX_E1_LC_HDB3;
        sE1CfgData.eRxLineCode                = OTX_E1_LC_HDB3;
        sE1CfgData.eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
        sE1CfgData.eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
        sE1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
        sE1CfgData.bfSiBits                   = 0x3;
        sE1CfgData.bfSaBits                   = 0x1F;
        sE1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
        sE1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
        sE1CfgData.bAutoResync                = OTX_TRUE;
        sE1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
        sE1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
        sE1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
        sE1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
        sE1CfgData.nLineLength                = 100;
        sE1CfgData.bHighTxPower               = OTX_FALSE;
        sE1CfgData.bReceiveEqualizer          = OTX_FALSE;;
        
        // Configure Transceivers
        for (i = 0; i < 2; i++) {
          // Set termination for the T1/E1 Front End
          result = OtxT1E1SetLineTermination(hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(hLi[i], result);
          // E1 mode
          result = OtxE1SetConfiguration(hLi[i], &sE1CfgData);  DemoCheckRc(hLi[i], result);
          result = OtxT1E1SetLiMode(hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(hLi[i], result);
          
          if (g_demoEnv.bUseExternalLoop) {
            // Use external loop cables
            result = OtxT1E1SetLineLoop(hLi[i], OTX_T1E1_LOOPS_OFF); DemoCheckRc(hLi[i], result);
          } else {
            // Tell the E1T1 device to loop its internal highway back to the internal highway
            result = OtxT1E1SetLineLoop(hLi[i], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(hLi[i], result);
          }
          
          // Although the timeslots connections are hardcoded on this card, we still need to connect
          // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
          // to the Li devices then it will send an Idle code on each timeslot.
          
          memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
          
          result = OtxDrvCreatePipe(64*32, &hPipe); DemoCheckRc(hPipe, result);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(hPipe, result);
          
          result = OtxDrvEnable(hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(hLi[i], result);
        }

      }
      
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif

    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId < 0 || nSourceId > (OTX_UINT32)g_demoEnv.coHighways) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

START:
  if (g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]));

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS
      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = fread(
        &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf,
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );
      
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32]; // On some boards this is bigger than we actually use
      int cbSingleTsBuf;
      int byteCount = 0;
      int cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / g_demoEnv.nHwyTimeslots;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        cbMaxSingleTsBuf,
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (cbSingleTsBuf < cbMaxSingleTsBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          DemoLogf("Pad with 0xFF [%d] at the end of the buffer to make sure that we don't underrun the Burst Sender\n", (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]));
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]) );
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        //DemoLogf("g_demoEnv.coHighways=%d, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend=%d \n", g_demoEnv.nCurrentHwy, g_demoEnv.coLi, g_demoEnv.coHighways, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend);
        for ( x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend; 
        (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
        x += g_demoEnv.nHwyTimeslots
          ) {
          g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
        }
        
        assert(byteCount == cbSingleTsBuf);
          
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * g_demoEnv.nHwyTimeslots;

    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: ui32 (%d) != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf (%d)\n", nSourceId, ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } else {
      
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        goto START;
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
      }
    }
    
  } else {
    // Send a gamma pattern, which means that we increase the value by one for each frame (continuously until aborted)
    int x, y;

    for (x = 0; x < DEMO_BURST_BUF_SIZE; x += g_demoEnv.nHwyTimeslots) {
      for (y = 0; y < g_demoEnv.nHwyTimeslots; y++) {
        g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[y]++;
        g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x+y] = g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[y];
      }
    }

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
    if(ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
      DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer.\n", nSourceId);
      DemoLogf("   Buffer size sent %d\n", g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      DemoLogf("   Buffer size returned %d\n", ui32);
    }
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // Echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // Note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("Burst Receiver Hwy[%d][%d] (len=%.5d) ID:%08x ", nSourceId, g_demoEnv.nCurrentHwy, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < g_demoEnv.nHwyTimeslots) && (x < (OTX_INT32) sizeof(buf)); x++) {
      DemoLogf("%.2X ", buf[x]);
    }

    DemoLogf("\n");
  }

  // Check the received value if we are in Gamma Test Mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
    int x;
    int y;
    
    for (x = 0; x < sizeof(buf); x += g_demoEnv.nHwyTimeslots) {
      for (y = 0; y < g_demoEnv.nHwyTimeslots; y++) {
        if (buf[x+y] == g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]) {
          // Data matched. No action needed.
        } else {
          // Data mismatch. Step error count and resync
          g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y] = buf[x+y];
          g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[y]++;
        }
        
        // Step the GammaTestData value, which is the next value that we expect
        g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]++;
      }
    }
  }

  
  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < (OTX_INT32) sizeof(buf); 
          x += g_demoEnv.nHwyTimeslots
      ) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      } else {
        //DemoLogf("Burst Receiver Hwy[%d] (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from a T1E1 device
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the Demo2.out logical device
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8196];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8196, &nBytesReturned, &nIoControlCode);
      if (otxRc == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO2_NOTIFY_DEBUG_DATA:
          {
            Demo2NotifyGetDebugDataS *pDataS = (Demo2NotifyGetDebugDataS *)buf;
            OTX_INT32 i;

            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            if (nBytesReturned != sizeof(Demo2NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("DEMO2_NOTIFY_DEBUG_DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);
              DemoLogf("nData3 = %.4X\n", pDataS->nData3);
              DemoLogf("nData4 = %.4X\n", pDataS->nData4);
              DemoLogf("nData5 = %.4X\n", pDataS->nData5);
              DemoLogf("nData6 = %.4X\n", pDataS->nData6);
              DemoLogf("nData7 = %.4X\n", pDataS->nData7);

              for (i = 0; i < 32; i++) {
                if ((i != 0) && (i%8==0)) DemoLogf("\n");
                DemoLogf("%.2d=%.4X ", i, (unsigned short)pDataS->sqBuf[i]);
              }
              DemoLogf("\n");

            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }
      }

    } else {

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Burst Highway. Currently %d is selected.\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" a   Abort Burst Hwy[%d] Senders and Receivers\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" r   Start Burst Receiver Hwy[%d] (no file logging)\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" R   Start Burst Receiver Hwy[%d] Prompt for file to recv\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" s   Start Burst Sender Hwy[%d]. Continously send a gamma pattern\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" S   Start Burst Sender Hwy[%d]. Prompt for file to send\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" L   Burst Sender Hwy[%d]. Toggle loop file sends (currently %s)\n", g_demoEnv.nCurrentHwy, g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop ? "On" : "Off");
    DemoLogf(" e   Enable Burst  (affects all burst highways)\n");
    DemoLogf(" d   Disable Burst (affects all burst highways)\n");
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" V   Toggle verbose (print send progress) setting Burst Sender[%d]\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" g   Enable/Disable Gamma Test\n");
    DemoLogf(" G   Clear Gamma Test error counters\n");
    DemoLogf(" j   Print Gamma Test Results for current H(condensed form)\n");
    DemoLogf(" J   Print Gamma Test Results (verbose form)\n");
    break;

  case 'g':
    {
      int nActiveBurstReceivers = 0;
      int nHwy = g_demoEnv.nCurrentHwy;
      
      if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
        // Burst receiver is running. Ok.

        if (!g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
          // Gamma mode is not enabled. So enable it.
          int n;
          memset(g_hLogicDevs.mBurstReceiver[nHwy].sqGammaTestData, 0, g_demoEnv.nHwyTimeslots);
          for (n = 0; n < g_demoEnv.nHwyTimeslots; n++) {
            g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
          }
          
          g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_TRUE;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now enabled\n", nHwy);

        } else {
          // Disable Gamma mode.
          g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_FALSE;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now disabled\n", nHwy);
        }
      } else {
        DemoLogf("Burst Receiver Hwy[%d] The Burst Receiver is not yet running. Start with 'r' or 'R'.\n");
      }
    }
    break;

  case 'G':  // Zero Gamma Test Error Counters
    {
      int n;
      for (n = 0; n < g_demoEnv.nHwyTimeslots; n++) {
        g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n] = 0;
      }
      
      DemoLogf("Burst Receiver Hwy[%d] Gamma Test Errors are now reset to zero\n", g_demoEnv.nCurrentHwy);
    }
    break;

  case 'j':  // Gamma Test Results (condensed form)
    {
      if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bActive == OTX_TRUE) {
        int y;
        if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bGammaTestMode) {
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results (Condensed Form):\n", g_demoEnv.nCurrentHwy);
          for (y = 0; y < g_demoEnv.nHwyTimeslots; y++) {
            if ((y % 32) == 0) {
              DemoLogf("TS[%.3d] ", y);
            }
            if ((y % 4) == 0) {
              DemoLogf(" "); // Group by 4 timeslots to make it easier to read
            }
            if ((y % 8) == 0) {
              DemoLogf(" "); // Group by 8 timeslots to make it easier to read
            }
            if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y] > 9) {
              DemoLogf("*"); // Indicates a number larger than 9
            } else {
              DemoLogf("%1d", g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y]);
            }
            if (((y+1) % 32) == 0) {
              DemoLogf("\n");
            }
          }
          DemoLogf("\n");
        } else {
          DemoLogf("Gamma Test Mode is not enabled\n");
        }
      } else {
        DemoLogf("Burst Receiver Threads are not running. Start them with 'r' or 'R'\n");
      }
    }
    break;

  case 'J':  // Gamma Test Results (verbose form)
    {
      if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bActive == OTX_TRUE) {
        if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bGammaTestMode) {
          int y;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results (Verbose Form):\n", g_demoEnv.nCurrentHwy);
          for (y = 0; y < g_demoEnv.nHwyTimeslots; y++) {
            DemoLogf("Hwy[%.1d] TS%.3d:  ", g_demoEnv.nCurrentHwy, y);
            DemoLogf("%d", g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y]);
            DemoLogf("\n");
          }
        } else {
          DemoLogf("Gamma Test Mode is not enabled\n");
        }
      } else {
        DemoLogf("Burst Receiver Threads are not running. Start them with 'r' or 'R'\n");
      }
    }
    break;



  case 'e':
    // by enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 'd':
    // by disabling the Burst device, we stop the process of transferring data
    // from the hardware to the driver via DMA.
    DemoLogf("Disable g_hPhysDevs.hBurst\n");
    result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 'v':
    g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose = !g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentHwy);
    break;

  case 'V':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentHwy);
    break;

  case 'a':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      }
      g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_FALSE;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("mBurstSender[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile);
        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;
      }
      g_hLogicDevs.mBurstSender[nHighwayNo].bActive   = OTX_FALSE;
    }
    break;

  case 'r':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", nHighwayNo);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("Burst Receiver Hwy[%d]: Closing file, but continuing to receive data.\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      } else {
        DemoLogf("Burst Receiver Hwy[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      char sqzFileName[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE || g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile == NULL) {
        DemoLogf("Burst Receiver Hwy[%d]. Please enter file to receive: ", nHighwayNo);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        } 

        strcpy(sqzFileName, sqzBuf);

        if ((g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = fopen(sqzFileName, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("Burst Receiver Hwy[%d]. Please enter single TS[0-%d] to receive to file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.nHwyTimeslots, nTsToRecord);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nTsToRecord = nTsToRecord;

        DemoLogf("Burst Receiver Hwy[%d]. Receiving data to file \"%s\". Abort using command 'a'.\n", nHighwayNo, sqzFileName);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else {
        DemoLogf("Burst Receiver Hwy[%d] Thread already running.\n", nHighwayNo);
      }
    }
    break;

  case 's':
    {
      OTX_UINT32 nHighwayNo, i;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Sender Hwy[%d]. Sending a gamma pattern continuously\n", nHighwayNo);

        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;

        // The starting point for the gamma pattern will be value=timeslot
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = 0;
        for (i = 0; i < (OTX_UINT32)g_demoEnv.nHwyTimeslots; i++) {
          g_hLogicDevs.mBurstSender[nHighwayNo].sqGammaData[i] = (unsigned char) i;
        }

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;

        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = DEMO_BURST_BUF_SIZE;
      } else {
        DemoLogf("Burst Sender Hwy[%d] already running.\n", nHighwayNo);
      }
    }

    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToSend = -1;

      nHighwayNo = g_demoEnv.nCurrentHwy;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Sender Hyw[%d]. Please enter file to send: ", nHighwayNo);

        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = fopen(sqzBuf, "rb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("Burst Sender Hwy[%d]. Please enter single TS[0-%d] to send from file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.nHwyTimeslots, nTsToSend);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToSend = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].nTsToSend = nTsToSend;

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;
        g_hLogicDevs.mBurstSender[nHighwayNo].bVerbose = OTX_TRUE;
      } else {
        DemoLogf("Burst Sender Hwy[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'L':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop;
    DemoLogf("Setting loop file sends %s for Burst Highway[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop) ? "ON" : "OFF", g_demoEnv.nCurrentHwy);
    break;


  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      Demo2IoOutGetFullFrameS dataS;

      if (g_hUserApp != OTX_INVALID_HANDLE_VALUE) {
        
        DemoLogf("Full Frame sample:\n");
        
        result = OtxDspIoControl(g_hUserApp, DEMO2_IOCTL_GET_FULL_FRAME, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hUserApp, result);
        
        if (OTX_SUCCESS(result)) {
          if (nBytesReturned != sizeof(Demo2IoOutGetFullFrameS)) {
            DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_FULL_FRAME (nBytesReturned=%d)\n", nBytesReturned);
          } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
            DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
          } else {
            OTX_INT32 i;
            for (i = 0; i < (OTX_INT32)g_nDspHighwaySpeed/64; i++) {
              if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
              DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
            }
            DemoLogf("\n");
            
          }
        }
      }

    }
    break;


  case 'M':
    {
      DemoLogf("Dsp Mem Test\n");
      OtxDrvDbgDump(g_hDsp, 4);
    }
    break;

  case 'F':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", 0);

      result = OtxDspIoControl(g_hUserApp, DEMO2_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, 0x1234, 5000); DemoCheckRc(g_hUserApp, result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;



  case '0':
  case '1':
    if (commandChar - '0' < g_demoEnv.coHighways) {
      // Select current Li
      g_demoEnv.nCurrentHwy = commandChar - '0';
      DemoLogf("HWY#%d Selected ....\n", g_demoEnv.nCurrentHwy);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coHighways        = 2;
  envP->nHwyTimeslots     = 128;
  envP->bDisableBurstTx   = OTX_FALSE;
  envP->bDisableBurstRx   = OTX_FALSE;

  envP->bUseExternalLoop = OTX_FALSE;
  envP->bUseDspLoop      = OTX_FALSE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // Use External loop with loop cables
          envP->bUseExternalLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'P') {   // Use loop via DSP
          envP->bUseDspLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coHighways = 1;
            envP->nHwyTimeslots = 128;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coHighways = 2;
            envP->nHwyTimeslots = 128;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coHighways = 2;
            envP->nHwyTimeslots = 32;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coHighways = 2;
            envP->nHwyTimeslots = 32;
          } else if ( *(argv[i]+2) == '1') {
            envP->nBoardType = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->coHighways = 1;
            envP->nHwyTimeslots = 32;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coHighways = 1;
            envP->nHwyTimeslots = 128;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|E|4|P|1|C] : Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  1 = Use Thor-1-PCI-Plus Board\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  DemoLogf(" /D[T|R]      : Disable one direction of the Burst device\n");
  DemoLogf("                  T = Disable Burst TX\n");
  DemoLogf("                  R = Disable Burst RX\n");
  DemoLogf(" /P           : Use loop via DSP \n");
  DemoLogf(" /X           : Use external loop cables (default is internal looping)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile,   formatStr, argList);

    fflush(g_pDemoLogFile);
  }
  va_end(argList);

}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

  if (g_pDemoLogFile) {
    fclose(g_pDemoLogFile);
  }
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("T8pBurstBasic, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_H100_ETCF, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\n");
  DemoLogf("Press 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Abort threads
  DemoLogf("\n");
  DemoLogf("Exiting...\n");
  {
    OTX_INT32 i;

    g_bStatsThreadActive = OTX_FALSE;
    g_bReadBurstStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
